/******* fermion_force_hisq_qop_P.c ****************/
/* MIMD version 7 */

/* This is the MILC wrapper for the SciDAC Level 3 QOP fermion force routine */
/* 11/2005 C. DeTar and D. Renner */
/* 5/09/07 C. DeTar Support selected precision */
/* NOTE: This is an include file for fermion_force_hisq_qop_F.c and
   fermion_force_hisq_qop_D.c.  Any changes must take this usage
   into account */

/* External entry points in this file are private to
   fermion_force_hisq_qop.c

   eo_fermion_force_oneterm_[FD]
   eo_fermion_force_twoterms_[FD]
   fermion_force_hisq_multi_[FD]
   fermion_force_hisq_block_[FD]

*/

/*
 * $Log: fermion_force_hisq_qop_P.c,v $
 *
 */

#if (QOP_Precision==1)

#define CREATE_F_FROM_SITE4        create_F_F_from_site4
#define CREATE_G_FROM_SITE4        create_F_G_from_site4
#define CREATE_V_FROM_SITE         create_F_V_from_site
#define CREATE_V_FROM_FIELD        create_F_V_from_field
#define EO_FERMION_FORCE_ONETERM   eo_fermion_force_oneterm_F
#define EO_FERMION_FORCE_TWOTERMS  eo_fermion_force_twoterms_F
#define FERMION_FORCE_HISQ_MULTI fermion_force_hisq_multi_F
#define FERMION_FORCE_HISQ_BLOCK fermion_force_hisq_block_F
#define MY_REAL                    float
#define LOAD_QOP_HISQ_COEFFS     load_qop_F_hisq_coeffs
#define UNLOAD_F_TO_SITE4          unload_F_F_to_site4
#define UNLOAD_G_TO_SITE4          unload_F_G_to_site4
				   
#else				   
				   
#define CREATE_F_FROM_SITE4        create_D_F_from_site4
#define CREATE_G_FROM_SITE4        create_D_G_from_site4
#define CREATE_V_FROM_SITE         create_D_V_from_site
#define CREATE_V_FROM_FIELD        create_D_V_from_field
#define EO_FERMION_FORCE_ONETERM   eo_fermion_force_oneterm_D
#define EO_FERMION_FORCE_TWOTERMS  eo_fermion_force_twoterms_D
#define FERMION_FORCE_HISQ_MULTI fermion_force_hisq_multi_D
#define FERMION_FORCE_HISQ_BLOCK fermion_force_hisq_block_D
#define MY_REAL                    double
#define LOAD_QOP_HISQ_COEFFS     load_qop_D_hisq_coeffs
#define UNLOAD_F_TO_SITE4          unload_D_F_to_site4
#define UNLOAD_G_TO_SITE4          unload_D_G_to_site4
#endif

#include "generic_ks_includes.h"
#include "../include/generic_qop.h"
#include "../include/generic_ks_qop.h"

/* Set default if undeclared */
#ifndef KS_MULTIFF
#define KS_MULTIFF FNMAT
#endif

#ifdef FFTIME
static const char *qop_prec[2] = {"F", "D"};
#endif

static char* cvsHeader = "$Header: /u/aquila/cvsroot/milc_qcd/generic_ks/fermion_force_hisq_qop_P.c,v 1.2 2007/12/14 04:36:31 detar Exp $";

/**********************************************************************/
/* Standard MILC interface for the single-species Hisq fermion force
   routine */
/**********************************************************************/
void EO_FERMION_FORCE_ONETERM( Real eps, Real weight, field_offset x_off,
			       ferm_links_t *fn, ks_action_paths *ap)
{

  /* For example weight = nflavors/4 */

  QOP_GaugeField *links;
  QOP_Force *mom;
  QOP_ColorVector *vecx;
  
  QOP_hisq_coeffs_t coeff;
  QOP_info_t info;

  double remaptime = -dclock();

  QOP_opt_t qop_ff_opt[2] = {
    {.tag = "fnmat_src_min",.value=4},
    {.tag = "veclength",.value=4}
  };

  /* Initialize QOP */
  if(initialize_qop() != QOP_SUCCESS){
    node0_printf("eo_fermion_force_oneterm: Error initializing QOP\n");
    terminate(1);
  }

  /* Convert gauge links and momentum to QOP format */

  links = CREATE_G_FROM_SITE4(F_OFFSET(link), EVENANDODD);
  mom   = CREATE_F_FROM_SITE4(F_OFFSET(mom), EVENANDODD);

  /* Convert color vector to QOP format */
  vecx = CREATE_V_FROM_SITE(x_off,EVENANDODD);

  /* Load coefficients */
  LOAD_QOP_HISQ_COEFFS(&coeff, weight, ap->act_path_coeff);

  /* Compute fermion force */
  remaptime += dclock();
  QOP_hisq_force_set_opts(qop_ff_opt, 2);
  QOP_hisq_force(&info, links, mom, &coeff, eps, vecx);
  remaptime -= dclock();

  /* Unload momentum */
  UNLOAD_F_TO_SITE4( F_OFFSET(mom), mom, EVENANDODD );

  /* Clean up */
  QOP_destroy_G(links); links = NULL;
  QOP_destroy_F(mom);   mom = NULL;
  QOP_destroy_V(vecx);  vecx = NULL;

  remaptime += dclock();
#ifdef FFTIME
  node0_printf("FFTIME:  time = %e (qop) terms = 1 mflops = %e\n",
	       info.final_sec, (Real)info.final_flop/(1e6*info.final_sec) );
#ifdef REMAP
  node0_printf("FFREMAP:  time = %e\n",remaptime);
#endif
#endif
}

/**********************************************************************/
/* Standard MILC interface for the two-species Hisq fermion force
   routine */
/**********************************************************************/
void EO_FERMION_FORCE_TWOTERMS( Real eps, Real weight1, Real weight2, 
				field_offset x1_off, field_offset x2_off,
				ferm_links_t *fn, ks_action_paths *ap )
{

  /* For example weight1 = nflavor1/4; weight2 = nflavor2/4 */

  QOP_GaugeField *links;
  QOP_Force *mom;
  QOP_ColorVector *vecx[2];
  
  QOP_hisq_coeffs_t coeff;
  MY_REAL epsv[2];
  QOP_info_t info;
  QOP_opt_t qop_ff_opt[2] = {
    {.tag = "fnmat_src_min",.value=4},
    {.tag = "veclength",.value=4}
  };

  double remaptime = -dclock();

  /* Initialize QOP */
  if(initialize_qop() != QOP_SUCCESS){
    printf("eo_fermion_force_twoterms: Error initializing QOP\n");
    terminate(1);
  }

  /* Convert gauge links and momentum to QOP format */

  links = CREATE_G_FROM_SITE4(F_OFFSET(link), EVENANDODD);
  mom   = CREATE_F_FROM_SITE4(F_OFFSET(mom), EVENANDODD);

  /* Loop over fermion species */
  /* Convert color vectors in site structure to QOP format */
  vecx[0] = CREATE_V_FROM_SITE(x1_off,EVENANDODD);
  vecx[1] = CREATE_V_FROM_SITE(x2_off,EVENANDODD);

  /* Load coefficients */
  epsv[0] = eps*weight1;  epsv[1] = eps*weight2;
  LOAD_QOP_HISQ_COEFFS(&coeff, 1., ap->act_path_coeff);

  /* Compute fermion force */
  remaptime += dclock();
  QOP_hisq_force_set_opts(qop_ff_opt, 2);
  QOP_hisq_force_multi(&info, links, mom, &coeff, epsv, vecx, 2);
  remaptime -= dclock();

  /* Unload momentum */
  UNLOAD_F_TO_SITE4( F_OFFSET(mom), mom, EVENANDODD );

  /* Clean up */
  QOP_destroy_G(links); links = NULL;
  QOP_destroy_F(mom);   mom = NULL;
  QOP_destroy_V(vecx[0]);  vecx[0] = NULL;
  QOP_destroy_V(vecx[1]);  vecx[1] = NULL;

  remaptime += dclock();
#ifdef FFTIME
  node0_printf("FFTIME:  time = %e (qop) terms = 2 mflops = %e\n",
	       info.final_sec, (Real)info.final_flop/(1e6*info.final_sec) );
#ifdef REMAP
  node0_printf("FFREMAP:  time = %e\n",remaptime);
#endif
#endif
}


/**********************************************************************/
/*   Version for hisq.  Parallel transport nterms source vectors    */
/**********************************************************************/

void FERMION_FORCE_HISQ_MULTI( Real eps, Real *residues, 
				 su3_vector **xxx, int nterms,
				 ferm_links_t *fn, ks_action_paths *ap ) 
{

  QOP_GaugeField *links;
  QOP_Force *mom;
  QOP_ColorVector **vecx;
  
  QOP_hisq_coeffs_t coeff;
  int i;
  MY_REAL *epsv;
  QOP_info_t info;
  char myname[] = "fermion_force_hisq_multi";

  double remaptime = -dclock();

  /* Initialize QOP */
  if(initialize_qop() != QOP_SUCCESS){
    printf("%s(%d): Error initializing QOP\n",myname,this_node);
    terminate(1);
  }

  /* Convert gauge links and momentum to QOP format */

  links = CREATE_G_FROM_SITE4(F_OFFSET(link), EVENANDODD);
  mom   = CREATE_F_FROM_SITE4(F_OFFSET(mom), EVENANDODD);

  /* Make space for QOP vector pointers */
  vecx = (QOP_ColorVector **)malloc(sizeof(QOP_ColorVector *)*nterms);
  if(vecx == NULL){
    printf("%s(%d): no room for vecx pointers\n",myname,this_node);
    terminate(1);
  }

  /* Convert color vectors to QOP format */
  for(i = 0; i < nterms; i++){
    vecx[i] = CREATE_V_FROM_FIELD(xxx[i], EVENANDODD);
  }

  /* Make space for weights */
  epsv = (MY_REAL *)malloc(sizeof(MY_REAL)*nterms);
  /* Load coefficients */
  for(i = 0; i < nterms; i++) epsv[i] = eps*residues[i];
  LOAD_QOP_HISQ_COEFFS(&coeff, 1., ap->act_path_coeff);

  /* Compute fermion force */
  remaptime += dclock();
  QOP_hisq_force_multi(&info, links, mom, &coeff, epsv, vecx, nterms);
  remaptime -= dclock();

  /* Unload momentum */
  UNLOAD_F_TO_SITE4( F_OFFSET(mom), mom, EVENANDODD );

  /* Clean up */
  QOP_destroy_G(links); links = NULL;
  QOP_destroy_F(mom);   mom = NULL;
  for(i = 0; i < nterms; i++){
    QOP_destroy_V(vecx[i]);  vecx[i] = NULL;
  }
  free(vecx);  vecx = NULL;
  free(epsv);  epsv = NULL;

  remaptime += dclock();
#ifdef FFTIME
  node0_printf("FFTIME:  time = %e (qop %s) terms = %d mflops = %e\n",
	       info.final_sec, qop_prec[QOP_Precision-1],
	       nterms, info.final_flop/(1e6*info.final_sec) );
#ifdef REMAP
  node0_printf("FFREMAP:  time = %e\n",remaptime);
#endif
#endif

}


/**********************************************************************/
/*   Parallel transport vectors in blocks of veclength. Hisq only   */
/**********************************************************************/
/* Requires the xxx1 and xxx2 terms in the site structure */

void FERMION_FORCE_HISQ_BLOCK( Real eps, Real *residues, 
				 su3_vector **xxx, int nterms, int veclength,
				 ferm_links_t *fn, ks_action_paths *ap ) {

  int i,j;
  site *s;

  /* First do blocks of size veclength */
  for( j = 0;  j <= nterms-veclength; j += veclength )
    FERMION_FORCE_HISQ_MULTI( eps, &(residues[j]), xxx+j, veclength, fn, ap );
  
#ifndef ONEMASS
  /* Continue with pairs if needed */
  for( ; j <= nterms-2 ; j+=2 ){
    FORALLSITES(i,s){
      s->xxx1 = xxx[j  ][i] ;
      s->xxx2 = xxx[j+1][i] ;
    }
    EO_FERMION_FORCE_TWOTERMS( eps, residues[j], residues[j+1],
			       F_OFFSET(xxx1), F_OFFSET(xxx2), fn, ap );
  }

  /* Finish with a single if needed */
  for( ; j <= nterms-1; j++ ){
    FORALLSITES(i,s){ s->xxx1 = xxx[j][i] ; }
    EO_FERMION_FORCE_ONETERM( eps, residues[j], F_OFFSET(xxx1), fn, ap );
  }
#else
  /* Thrown in just to make it compile.  Probably never used. */
  /* Finish with a single if needed */
  for( ; j <= nterms-1; j++ ){
    FORALLSITES(i,s){ s->xxx = xxx[j][i] ; }
    EO_FERMION_FORCE_ONETERM( eps, residues[j], F_OFFSET(xxx), fn, ap );
  }
#endif

}

